# pypcapfile.savefile.py
"""
Core functionality for reading and parsing libpcap savefiles. This contains
the core classes pcap_packet and pcap_savefile, as well as the core function
load_savefile.
"""

import binascii
import ctypes
import struct
import sys

import pcapfile.linklayer as linklayer

from pcapfile.structs import __pcap_header__, pcap_packet

VERBOSE = False

_MAGIC_NUMBER = 0xa1b2c3d4
_MAGIC_NUMBER_NS = 0xa1b23c4d


def __TRACE__(msg, args=None):
    if VERBOSE:
        if args:
            print(msg.format(*args))
        else:
            print(msg)


class pcap_savefile(object):
    """
    Represents a libpcap savefile. The packets member is a list of pcap_packet
    instances. The 'valid' member will be None for an uninitialised instance,
    False if the initial validation fails, or True if the instance has been
    successfully set up and the file has been parsed.
    """
    def __init__(self, header, packets=None):
        if not packets:
            packets = []
        self.header = header
        self.packets = packets
        self.valid = None
        self.byteorder = sys.byteorder

        if not self.__validate__():
            self.valid = False
        else:
            self.valid = True

        assert self.valid, 'Invalid savefile.'

    def __validate__(self):
        assert __validate_header__(self.header), "Invalid header."
        if not __validate_header__(self.header):
            return False

        # TODO: extended validation
        valid_packet = lambda pkt: (pkt is not None or
                                    pkt.issubclass(ctypes.Structure))
        if not 0 == len(self.packets):
            valid_packet = [valid_packet(pkt) for pkt in self.packets]
            assert False not in valid_packet, 'Invalid packets in savefile.'
            if False in valid_packet:
                return False

        return True

    def __repr__(self):
        string = '%s-endian capture file version %d.%d\n'
        string += '%ssecond time resolution\n'
        string += 'snapshot length: %d\n'
        string += 'linklayer type: %s\nnumber of packets: %d\n'
        return string % (self.header.byteorder, self.header.major,
                         self.header.minor,
                         "nano" if self.header.ns_resolution else "micro",
                         self.header.snaplen,
                         linklayer.lookup(self.header.ll_type),
                         len(self.packets))


def _load_savefile_header(file_h):
    """
    Load and validate the header of a pcap file.
    """
    try:
        raw_savefile_header = file_h.read(24)
    except UnicodeDecodeError:
        print("\nMake sure the input file is opened in read binary, 'rb'\n")
        raise

    # in case the capture file is not the same endianness as ours, we have to
    # use the correct byte order for the file header
    if raw_savefile_header[:4] in [struct.pack(">I", _MAGIC_NUMBER),
                                   struct.pack(">I", _MAGIC_NUMBER_NS)]:
        byte_order = b'big'
        unpacked = struct.unpack('>IhhIIII', raw_savefile_header)
    elif raw_savefile_header[:4] in [struct.pack("<I", _MAGIC_NUMBER),
                                     struct.pack("<I", _MAGIC_NUMBER_NS)]:
        byte_order = b'little'
        unpacked = struct.unpack('<IhhIIII', raw_savefile_header)
    else:
        raise Exception('Invalid pcap file.')

    (magic, major, minor, tz_off, ts_acc, snaplen, ll_type) = unpacked
    header = __pcap_header__(magic, major, minor, tz_off, ts_acc, snaplen,
                             ll_type, ctypes.c_char_p(byte_order),
                             magic == _MAGIC_NUMBER_NS)
    if not __validate_header__(header):
        raise Exception('invalid savefile header!')
    else:
        return header


def load_savefile(input_file, layers=0, verbose=False):
    """
    Parse a savefile as a pcap_savefile instance. Returns the savefile
    on success and None on failure. Verbose mode prints additional information
    about the file's processing. layers defines how many layers to descend and
    decode the packet. input_file should be a Python file object.
    """
    global VERBOSE
    old_verbose = VERBOSE
    VERBOSE = verbose

    __TRACE__('[+] attempting to load {:s}', (input_file.name,))

    header = _load_savefile_header(input_file)
    if __validate_header__(header):
        __TRACE__('[+] found valid header')
        packets = _load_packets(input_file, header, layers)
        __TRACE__('[+] loaded {:d} packets', (len(packets),))
        sfile = pcap_savefile(header, packets)
        __TRACE__('[+] finished loading savefile.')
    else:
        __TRACE__('[!] invalid savefile')
        sfile = None

    VERBOSE = old_verbose
    return sfile


def __validate_header__(header):
    if not type(header) == __pcap_header__:
        return False

    if header.magic not in [_MAGIC_NUMBER, _MAGIC_NUMBER_NS]:
        return False

    assert header.byteorder in [b'little', b'big'], 'Invalid byte order.'

    # as of savefile format 2.4, 'a 4-byte time zone offset; this
    # is always 0'; the same is true of the timestamp accuracy.
    if not header.tz_off == 0:
        return False

    if not header.ts_acc == 0:
        return False

    return True


def _load_packets(file_h, header, layers=0):
    """
    Read packets from the capture file. Expects the file handle to point to
    the location immediately after the header (24 bytes).
    """
    pkts = []

    hdrp = ctypes.pointer(header)
    while True:
        pkt = _read_a_packet(file_h, hdrp, layers)
        if pkt:
            pkts.append(pkt)
        else:
            break

    return pkts


def _read_a_packet(file_h, hdrp, layers=0):
    """
    Reads the next individual packet from the capture file. Expects
    the file handle to be somewhere after the header, on the next
    per-packet header.
    """
    raw_packet_header = file_h.read(16)
    if not raw_packet_header or len(raw_packet_header) != 16:
        return None

    # in case the capture file is not the same endianness as ours, we have to
    # use the correct byte order for the packet header
    if hdrp[0].byteorder == 'big':
        packet_header = struct.unpack('>IIII', raw_packet_header)
    else:
        packet_header = struct.unpack('<IIII', raw_packet_header)
    (timestamp, timestamp_ms, capture_len, packet_len) = packet_header
    raw_packet_data = file_h.read(capture_len)

    if not raw_packet_data or len(raw_packet_data) != capture_len:
        return None

    if layers > 0:
        layers -= 1
        raw_packet = linklayer.clookup(hdrp[0].ll_type)(raw_packet_data,
                                                        layers=layers)
    else:
        raw_packet = binascii.hexlify(raw_packet_data)

    packet = pcap_packet(hdrp, timestamp, timestamp_ms, capture_len,
                         packet_len, raw_packet)
    return packet



"""
Common structs for the package.
"""

import binascii
import ctypes


class __pcap_header__(ctypes.Structure):
    """
    C-struct representation of a savefile header. See __validate_header__
    for validation.
    """
    _fields_ = [('magic', ctypes.c_uint),          # file magic number
                ('major', ctypes.c_ushort),        # major version number
                ('minor', ctypes.c_ushort),        # minor version number
                ('tz_off', ctypes.c_uint),         # timezone offset
                ('ts_acc', ctypes.c_uint),         # timestamp accuracy
                ('snaplen', ctypes.c_uint),        # snapshot length
                ('ll_type', ctypes.c_uint),        # link layer header type
                ('byteorder', ctypes.c_char_p),    # byte order specifier
                ('ns_resolution', ctypes.c_bool)]  # nanosecond resolution


class pcap_packet(ctypes.Structure):
    """
    ctypes Structure representation of a packet. The header field is a pointer
    to the header of the savefile the packet came from to provide context. It
    can be accessed with header[0]. By default, the raw packet is stored in
    a string of hexadecimal-encoded bytes as the packet field. The raw()
    method will return the raw binary packet.
    """
    _fields_ = [('header', ctypes.POINTER(__pcap_header__)),
                ('timestamp', ctypes.c_uint),
                ('timestamp_ms', ctypes.c_uint),
                ('capture_len', ctypes.c_uint),
                ('packet_len', ctypes.c_uint)]

    packet = None

    def __init__(self, header, timestamp, timestamp_ms, capture_len,
                 packet_len, packet):
        super(pcap_packet, self).__init__()
        self.header = header
        self.timestamp = timestamp
        self.timestamp_ms = timestamp_ms
        self.capture_len = capture_len
        self.packet_len = packet_len
        self.packet = packet

    def raw(self):
        """
        Return the raw binary data from the packet.
        """
        return binascii.unhexlify(self.packet)

    def __repr__(self):
        if isinstance(self.packet, str):
            try:
                return self.raw()
            except TypeError:
                return str(self.packet)
        else:
            return str(self.packet)



"""
IP protocol definitions.
"""

import binascii
import ctypes
import struct


class Ethernet(ctypes.Structure):
    """
    Represents an Ethernet frame.
    """

    _fields_ = [('dst', ctypes.c_char_p),
                ('src', ctypes.c_char_p),
                ('type', ctypes.c_ushort)]

    payload = None

    def __init__(self, packet, layers=0):
        (dst, src, self.type) = struct.unpack('!6s6sH', packet[:14])

        dst = bytearray(dst)
        src = bytearray(src)
        self.dst = b':'.join([('%02x' % o).encode('ascii') for o in dst])
        self.src = b':'.join([('%02x' % o).encode('ascii') for o in src])

        payload = binascii.hexlify(packet[14:])
        self.payload = payload

        if layers:
            self.load_network(layers)

    def load_network(self, layers=1):
        """
        Given an Ethernet frame, determine the appropriate sub-protocol;
        If layers is greater than zerol determine the type of the payload
        and load the appropriate type of network packet. It is expected
        that the payload be a hexified string. The layers argument determines
        how many layers to descend while parsing the packet.
        """
        if layers:
            ctor = payload_type(self.type)[0]
            if ctor:
                ctor = ctor
                payload = binascii.unhexlify(self.payload)
                self.payload = ctor(payload, layers - 1)
            else:
                # if no type is found, do not touch the packet.
                pass

    def __str__(self):
        frame = 'ethernet from %s to %s type %s'
        frame = frame % (self.src, self.dst, payload_type(self.type)[1])
        return frame


def strip_ethernet(packet):
    """
    Strip the Ethernet frame from a packet.
    """
    if not type(packet) == Ethernet:
        packet = Ethernet(packet)
    payload = packet.payload

    if type(payload) == str:
        payload = binascii.unhexlify(payload)
    return payload


def payload_type(ethertype):
    """
    Returns the appropriate payload constructor based on the supplied
    EtherType.
    """
    if ethertype == 0x0800:
        from pcapfile.protocols.network.ip import IP
        return (IP, 'IPv4')
#    elif ethertype == 0x0806:
#        from pcapfile.protocols.network.arp import ARP
#        return (ARP, 'ARP')
#    elif ethertype == 0x0835:
#        from pcapfile.protocols.network.rarp import RARP
#        return (RARP, 'RARP')
#    elif ethertype == 0x08DD:
#        from pcapfile.protocols.network.ipv6 import IPv6
#        return (IPv6, 'IPv6')
    else:
        return (None, 'unknown')



#!/usr/bin/env python
"""
This is the test case for the savefile.
"""

import unittest

from pcapfile import savefile
from pcapfile import linklayer
from pcapfile.protocols.linklayer import ethernet


class TestCase(unittest.TestCase):
    """
    Validate the linklayer utility functions.
    """

    capfile = None

    def init_capfile(self, layers=0):
        """
        Initialise capture file.
        """
        self.capfile = savefile.load_savefile(open('test_data/test.pcap', 'r'),
                                              layers=layers)

    @classmethod
    def setUpClass(cls):
        """
        Print an intro to identify this test suite when running multiple tests.
        """
        print('[+] loading toplevel linklayer utility tests')

    def test_constructor_lookup(self):
        """
        Ensure the proper validation function is passed from the constructor
        lookup.
        """
        self.assertEqual(ethernet.Ethernet, linklayer.clookup(1))

    def test_lookup(self):
        """
        Test the strings returned by the short lookup functions.
        """
        self.assertEqual('LINKTYPE_ETHERNET', linklayer.lookup(1),
                         'invalid long name')
        self.assertEqual('ethernet', linklayer.slookup(1),
                         'invalid short name')



#!/usr/bin/env python
"""
This is the front end to the pcapfile test SUITE.
"""

import unittest


from pcapfile.test.linklayer_test import TestCase as LinklayerTest
from pcapfile.test.savefile_test import TestCase as SavefileTest
from pcapfile.test.protocols_linklayer_ethernet import TestCase as EthernetTest

if __name__ == '__main__':
    TEST_CLASSES = [SavefileTest, LinklayerTest, EthernetTest]
    SUITE = unittest.TestSuite()
    LOADER = unittest.TestLoader()
    for test_class in TEST_CLASSES:
        SUITE.addTests(LOADER.loadTestsFromTestCase(test_class))
    unittest.TextTestRunner(verbosity=2).run(SUITE)



#!/usr/bin/env python
"""
This is the test case for the savefile.
"""

import os
import pickle
import tempfile
import unittest
import base64
import sys

import pcapfile.test.fixture as fixture
from pcapfile import savefile


def create_pcap():
    """
    Create a capture file from the test fixtures.
    """
    tfile = tempfile.NamedTemporaryFile()
    if sys.version_info[0] >= 3:  # python3
        capture = pickle.loads(base64.b64decode(fixture.TESTPCAP3))
    else:  # python2 unsupported pickle protocol: 3
        capture = pickle.loads(fixture.TESTPCAP2.decode('base64'))
    open(tfile.name, 'wb').write(capture)
    return tfile


class TestCase(unittest.TestCase):
    """
    Test case for the savefile code.
    """
    capfile = None

    def init_capfile(self, layers=0):
        """Initialise the capture file."""
        tfile = create_pcap()
        self.capfile = savefile.load_savefile(tfile, layers=layers)
        tfile.close()
        if os.path.exists(tfile.name):
            os.unlink(tfile.name)

    @classmethod
    def setUpClass(cls):
        """
        Print an intro to identify this test suite when running multiple tests.
        """
        print('[+] loading basic tests')

    def setUp(self):
        """
        Set up a default capture file.
        """
        # only need to initialise capfile on the first time, while being able
        # load it with additional decoding layers.
        if not self.capfile:
            self.init_capfile()

    def test_network_load(self):
        """
        Test that the code that loads network layer packets from the
        top level works.
        """
        self.init_capfile(layers=2)
        for packet in self.capfile.packets:
            for field in ['src', 'dst', 'v', 'hl', 'tos', 'ttl']:
                ipkt = packet.packet.payload
                self.assertTrue(hasattr(ipkt, field), 'invalid packet!')

    def test_frame_load(self):
        """
        Ensure that ethernet frames load from the top level.
        """
        self.init_capfile(layers=1)
        for packet in self.capfile.packets:
            for field in ['src', 'dst', 'type', 'payload']:
                self.assertTrue(hasattr(packet.packet, field),
                                'invalid frame!')

    def test_packet_valid(self):
        """
        Make sure raw packets load properly.
        """
        packet = self.capfile.packets[0].raw()

        self.assertEqual(int(bytearray(packet)[14]), 69, 'invalid packet')

        for packet in self.capfile.packets:
            for field in ['capture_len', 'timestamp', 'timestamp_ms',
                          'packet', 'header', 'packet_len']:
                self.assertTrue(hasattr(packet, field), 'invalid packet!')

    def test_header_valid(self):
        """
        Test to ensure the header validation code works.
        """
        header = self.capfile.header
        self.assertEqual(header.major, 2, 'invalid major version!')
        self.assertEqual(header.minor, 4, 'invalid minor version!')

    def test_basic_import(self):
        """
        Validate basic parameters of a simple savefile load.
        """
        self.assertTrue(self.capfile.valid, 'invalid capture file')
        self.assertEqual(len(self.capfile.packets), 23,
                         'wrong number of packets!')



# contains definitions of link layer types from libpcap
"""
Link layer definitions and lookup functions based on libpcap's LLTYPE_*
values.
"""

import imp
import sys

import pcapfile.protocols.linklayer.ethernet as ethernet


__LL_TYPES__ = [('LINKTYPE_NULL', 0, 'null', None),
                ('LINKTYPE_ETHERNET', 1, 'ethernet',
                    ethernet.Ethernet),
                ('LINKTYPE_TOKEN_RING', 6, 'token ring', None),
                ('LINKTYPE_ARCNET', 7, 'ARCnet', None),
                ('LINKTYPE_SLIP', 8, 'SLIP', None)]


def __get_ll_type__(ll_type):
    """
    Given an lltype value, retrieve its definition.
    """
    res = [llt for llt in __LL_TYPES__
           if llt[1] == ll_type]
    assert len(res) < 2, 'Duplicate linklayer types.'

    if res:
        return res[0]
    else:
        return None


def lookup(ll_type):
    """
    Given an ll_type, retrieve the appropriate LL_TYPE.
    """
    res = __get_ll_type__(ll_type)
    if res:
        return res[0]
    else:
        return res


def slookup(ll_type):
    """
    Given an ll_type, retrieve the short name for the link layer.
    """
    res = __get_ll_type__(ll_type)
    if res:
        return res[2]
    else:
        return res


def clookup(ll_type):
    """
    Given an ll_type, retrieve the linklayer constructor to decode
    the packets.
    """
    res = __get_ll_type__(ll_type)
    if res:
        return res[3]
    else:
        return res


def __load_linktype__(link_type):
    """
    Given a string for a given module, attempt to load it.
    """

    try:
        filep, pathname, description = imp.find_module(link_type, sys.path)
        link_type_module = imp.load_module(link_type, filep, pathname,
                                           description)
    except ImportError:
        return None
    finally:
        if filep:
            filep.close()

    return link_type_module



#!/usr/bin/env python
"""
This is the test case for the savefile.
"""

import binascii
import unittest

from pcapfile.protocols.linklayer import ethernet

# a sample ethernet frame carrying an IPv4 SYN request
TEST_PACKET = [b'010203040506ffeeddccbbaa08004500003cc43800003806',
               b'2898adc22552c0a8022f0050f9dc06511a336489a374a012',
               b'37641b070000020405960402080a2195430027e3b0900103',
               b'0306']
TEST_PACKET = binascii.unhexlify(b''.join(TEST_PACKET))


class TestCase(unittest.TestCase):
    """
    Test case for the Ethernet frame parser.
    """

    @classmethod
    def setUpClass(cls):
        """
        Print a start message when loading the test suite.
        """
        print('[+] testing ethernet frame decoding...')

    def test_frame_field(self):
        """
        Verify attributes of Ethernet instance.
        """
        frame = ethernet.Ethernet(TEST_PACKET)
        self.assertEqual(frame.src, b'ff:ee:dd:cc:bb:aa',
                         'invalid frame source address')
        self.assertEqual(frame.dst, b'01:02:03:04:05:06',
                         'invalid frame destination address')
        self.assertEqual(frame.type, 0x0800, 'invalid frame type')
        self.assertTrue(hasattr(frame, 'load_network'),
                        'missing load_network method')
        frame.load_network()
        self.assertTrue(hasattr(frame.payload, 'ttl'),
                        'load_network fails')

    def test_frame_load(self):
        """
        Verify the ethernet frames load.
        """
        frame = ethernet.Ethernet(TEST_PACKET)
        self.assertTrue(isinstance(frame, ethernet.Ethernet),
                        'invalid ethernet frame!')



#!/usr/bin/env python

import os
from distutils.core import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='pypcapfile',
      version='0.10.0',
      description=('Pure Python package for reading and parsing libpcap '
                       'savefiles.'),
      long_description=read('README.rst'),
      author='Kyle Isom',
      author_email='coder@kyleisom.net',
      license='ISC',
      url='http://kisom.github.com/pypcapfile',
      scripts=['pcapfile_info',],
      packages=['pcapfile', 
                'pcapfile.test',
                'pcapfile.protocols',
                'pcapfile.protocols.linklayer',
                'pcapfile.protocols.network',
                ],
      package_data={'pcapfile.test': ['test/test_data']},
      data_files=[('share/doc/pcapfile',
                   ['README.rst', 'AUTHORS', 'CONTRIBUTING'])]
      )




"""
IP protocol definitions.
"""

import binascii
import ctypes
import struct


class IP(ctypes.Structure):
    """
    Represents an IP packet.
    """

    _fields_ = [('v', ctypes.c_ushort),         # version
                ('hl', ctypes.c_ushort),        # internet header length
                ('tos', ctypes.c_ubyte),        # type of service
                ('len', ctypes.c_ushort),       # total length
                ('id', ctypes.c_ushort),        # IPID
                ('flags', ctypes.c_ushort, 3),  # flags
                ('off', ctypes.c_ushort, 13),   # fragmentation offset
                ('ttl', ctypes.c_ubyte),        # TTL
                ('p', ctypes.c_ubyte),          # protocol
                ('sum', ctypes.c_ushort),       # checksum
                ('src', ctypes.c_char_p),       # source address
                ('dst', ctypes.c_char_p),       # destination address
                ('opt', ctypes.c_char_p),       # IP options
                ('pad', ctypes.c_char_p),       # padding bytes
                ('payload', ctypes.c_char_p)]   # packet payload

    def __init__(self, packet, layers=0):
        # parse the required header first, deal with options later
        magic = struct.unpack('!B',packet[0:1])[0]
        assert ((magic & 0b1100) == 4 and
                (magic & 0b0111) > 4), 'not an IPv4 packet.'

        fields = struct.unpack('!BBHHHBBHII', packet[:20])
        self.v = fields[0] & 0b1100
        self.hl = fields[0] & 0b0111
        self.tos = fields[1]
        self.len = fields[2]
        self.id = fields[3]
        self.flags = fields[4] >> 13
        self.off = fields[4] & 0x1fff
        self.ttl = fields[5]
        self.p = fields[6]
        self.sum = fields[7]
        self.src = ctypes.c_char_p(parse_ipv4(fields[8]))
        self.dst = ctypes.c_char_p(parse_ipv4(fields[9]))

        if self.hl > 0x14:
            start = 0x15
            end = self.len - 0x14
            self.opt = binascii.hexlify(packet[start:end])
        else:
            self.opt = b'\x00'
            self.payload = ctypes.c_char_p(binascii.hexlify(packet[0x14:]))

        self.pad = b'\x00'

    def __str__(self):
        packet = 'ipv4 packet from %s to %s carrying %d bytes'
        packet = packet % (self.src, self.dst, (len(self.payload) / 2))
        return packet


def parse_ipv4(address):
    """
    Given a raw IPv4 address (i.e. as an unsigned integer), return it in
    dotted quad notation.
    """
    raw = struct.pack('I', address)
    octets = struct.unpack('BBBB', raw)[::-1]
    ipv4 = b'.'.join([('%d' % o).encode('ascii') for o in bytearray(octets)])
    return ipv4


def strip_ip(packet):
    """
    Remove the IP packet layer, yielding the transport layer.
    """
    if not type(packet) == IP:
        packet = IP(packet)
    payload = packet.payload

    if type(payload) == '':
        payload = binascii.unhexlify(payload)
    return payload


def __call__(packet):
    return IP(packet)
